from typing import Any, Optional
from .._group import task_group
from ...task_input.str_input import StrInput
from ...task.resource_maker import ResourceMaker
from ...task.task import Task
from ...runner import runner
from ...helper.render_data.replacement_template import (
    Replacement
)
from ...helper.common import to_snake_case, coalesce
from ...helper.codemod.add_import_module import add_import_module
from ...helper.codemod.add_assert_module import add_assert_module

import os

# Common definitions

current_dir = os.path.dirname(__file__)

inputs = [
    StrInput(
        name='task-dir',
        prompt='Task directory',
        default='myTask'
    ),
    StrInput(
        name='task-name',
        prompt='Task name (can be empty)',
        default=''
    ),
]

replacements = Replacement().add_key_val(
    key='taskDir',
    value='input.task_dir'
).add_transformed_key_val(Replacement.ALL)(
    key='taskName',
    value=[
        'input.task_name',
        'os.path.basename(input.task_dir)'
    ]
).get()


def is_path_inside(path, parent_path):
    # Normalize the paths to avoid inconsistencies due to
    # different separators, etc.
    path = os.path.normpath(path)
    parent_path = os.path.normpath(parent_path)

    # Split the paths into components
    path_components = path.split(os.sep)
    parent_path_components = parent_path.split(os.sep)

    # Check if all components of `path` match the corresponding
    # components of `parent_path`
    return all(a == b for a, b in zip(path_components, parent_path_components))


def get_zrb_project_dir(project_dir: str) -> Optional[str]:
    if os.path.isfile(os.path.join(project_dir, 'zrb_init.py')):
        return project_dir
    new_project_dir = os.path.dirname(project_dir)
    if new_project_dir == project_dir:
        # No project found
        return None
    return get_zrb_project_dir(new_project_dir)


def _task_validate_create(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    input_map = task.get_input_map()
    task_dir = os.path.abspath(input_map.get('task_dir'))
    zrb_project_dir = get_zrb_project_dir(task_dir)
    if not zrb_project_dir:
        raise Exception(' '.join([
            f'Task directory: {task_dir}',
            'is not located in a project'
        ]))


def _task_create(*args: Any, **kwargs: Any):
    task_dir = os.path.abspath(kwargs.get('task_dir'))
    snake_task_name = to_snake_case(coalesce(
        kwargs.get('task_name'), os.path.basename(task_dir)
    ))
    zrb_project_dir = get_zrb_project_dir(task_dir)
    task_module_path = '.'.join(
        os.path.relpath(
            os.path.join(task_dir, snake_task_name),
            zrb_project_dir
        ).split(os.path.sep)
    )
    zrb_init_path = os.path.join(zrb_project_dir, 'zrb_init.py')
    with open(zrb_init_path, 'r') as f:
        code = f.read()
        code = add_import_module(code, task_module_path)
        code = add_assert_module(code, task_module_path)
    with open(zrb_init_path, 'w') as f:
        f.write(code)
    return True


# Task definitions

task_validate_create = Task(
    name='task-validate-create',
    inputs=inputs,
    runner=_task_validate_create
)

task_copy_resource = ResourceMaker(
    name='task-copy-resource',
    inputs=inputs,
    upstreams=[task_validate_create],
    replacements=replacements,
    template_path=os.path.join(current_dir, 'task_template'),
    destination_path='{{input.task_dir}}',
    scaffold_locks=['{{input.task_dir}}']
)

task_create = Task(
    name='task',
    group=task_group,
    inputs=inputs,
    runner=_task_create,
    upstreams=[task_copy_resource]
)
runner.register(task_create)



import libcst as cst


def add_assert_module(code: str, module_path: str) -> str:
    module = cst.parse_module(code)

    # Create the new assertion statement
    new_assert = cst.SimpleStatementLine([
        cst.Assert(
            test=cst.Name(value=module_path.split(".")[-1]),
            msg=None
        ),
    ])

    module.body.append(new_assert)

    # Generate the modified code
    modified_code = module.code
    return modified_code



from typing import Any
from ._group import md5_group
from ..task.task import Task
from ..task_input.str_input import StrInput
from ..runner import runner

import hashlib

# Common definitions

inputs = [
    StrInput(
        name='text',
        shortcut='t',
        description='Text',
        default=''
    )
]


def _md5_hash(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    hashed_text = hashlib.md5(text.encode()).hexdigest()
    return hashed_text


def _md5_sum(*args: str, **kwargs: Any):
    file_path: str = kwargs.get('file', '')
    with open(file_path, "rb") as f:
        contents = f.read()
        hashed_text = hashlib.md5(contents).hexdigest()
        return hashed_text


# Task definitions

md5_hash = Task(
    name='hash',
    group=md5_group,
    inputs=[
        StrInput(
            name='text',
            shortcut='t',
            description='Text',
            default=''
        )
    ],
    runner=_md5_hash,
    description='Encode md5 task',
    retry=0
)
runner.register(md5_hash)

md5_sum = Task(
    name='decode',
    group=md5_group,
    inputs=[
        StrInput(
            name='file',
            shortcut='f',
            description='File',
            default=''
        )
    ],
    runner=_md5_sum,
    description='Decode md5 task',
    retry=0
)
runner.register(md5_sum)



from typing import Any
import re


def coalesce(value: Any, *alternatives: Any) -> Any:
    if value is not None and value != '':
        return value
    for alternative in alternatives:
        if alternative is not None and value != '':
            return alternative
    return ''


def to_camel_case(string: str) -> str:
    pascal = to_pascal_case(string)
    if len(pascal) == 0:
        return pascal
    return pascal[0].lower() + pascal[1:]


def to_pascal_case(string: str) -> str:
    string = _to_alphanum(string)
    return ''.join([
        x.lower().capitalize() for x in to_human_readable(string).split(' ')
    ])


def to_kebab_case(string: str) -> str:
    string = _to_alphanum(string)
    return '-'.join([
        x.lower() for x in to_human_readable(string).split(' ')
    ])


def to_snake_case(string: str) -> str:
    string = _to_alphanum(string)
    return '_'.join([
        x.lower() for x in to_human_readable(string).split(' ')
    ])


def _to_alphanum(string: str) -> str:
    return re.sub(r'[^a-zA-Z0-9]+', ' ', string)


def to_human_readable(string: str) -> str:
    string = string.replace('-', ' ').replace('_', ' ')
    parts = string.split(' ')
    new_parts = []
    for part in parts:
        new_part = ''
        for char_index, char in enumerate(part):
            is_first = char_index == 0
            is_last = char_index == len(part) - 1
            previous_char = part[char_index - 1] if not is_first else ''
            next_char = part[char_index + 1] if not is_last else ''
            if char.isupper() and char != ' ' and (
                (not is_last and next_char.islower()) or
                (not is_first and previous_char.islower())
            ):
                new_part += ' ' + char
                continue
            new_part += char
        new_part = new_part.strip(' ')
        if new_part != '':
            new_parts.append(new_part)
    return ' '.join(new_parts).strip(' ')



from typing import Any, Mapping
from ._group import env_group
from ..helper.accessories.color import colored
from ..task.task import Task
from ..runner import runner

import os

# Common definitions


def _env_show(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    env_map = task.get_env_map()
    names = list(env_map.keys())
    names.sort()
    for name in names:
        value = os.getenv(name)
        colored_name = colored(name, color='green', attrs=['bold'])
        colored_equal = colored('=', color='grey', attrs=['dark'])
        colored_value = colored(value, attrs=['bold'])
        task.print_out(f'{colored_name}{colored_equal}{colored_value}')


def _env_get(*args: Any, **kwargs: Any) -> Mapping[str, str]:
    task: Task = kwargs['_task']
    return task.get_env_map()


# Task definitions

env_show = Task(
    name='show',
    group=env_group,
    runner=_env_show,
    description='Show environment values'
)
runner.register(env_show)

env_get = Task(
    name='get',
    group=env_group,
    runner=_env_get,
    description='get environment values'
)
runner.register(env_get)



from zrb.task.task import Task
from zrb.task_input.str_input import StrInput


def test_task_with_no_runner():
    task = Task(
        name='simple-no-error',
    )
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result


def test_task_with_predefined_runner():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        runner=_run
    )
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result == 'hello'


def test_task_with_decorated_runner():
    task = Task(
        name='hello',
    )

    @task.runner
    def _run(*args, **kwargs) -> str:
        return 'hello'

    main_loop = task.create_main_loop()
    result = main_loop()
    assert result == 'hello'


def test_task_with_predefined_and_decorated_runner():
    def _run(*args, **kwargs) -> str:
        return 'hello'
    task = Task(
        name='hello',
        runner=_run
    )
    # should throw error when add decorated runner
    # to a task with pre-existing runner
    is_error = False
    try:
        @task.runner
        def _run(*args, **kwargs) -> str:
            return 'not-hello'
    except Exception:
        is_error = True
    assert is_error
    main_loop = task.create_main_loop()
    result = main_loop()
    assert result == 'hello'


def test_task_with_input():
    def _run(*args, **kwargs) -> str:
        name = kwargs['name']
        favorite_drink = kwargs['favorite_drink']
        return f'hello {name}, your favorite drink is {favorite_drink}'
    task = Task(
        name='hello-name',
        inputs=[
            StrInput(name='name'),
            StrInput(name='favorite-drink')
        ],
        runner=_run
    )
    main_loop = task.create_main_loop()
    result = main_loop(name='Dumbledore', favorite_drink='Elixir')
    assert result == 'hello Dumbledore, your favorite drink is Elixir'



from .helper.cli import create_cli
from .helper.log import logger
from .helper.accessories.color import colored

logger.info(colored('Creating CLI instance', attrs=['dark']))
cli = create_cli()

if __name__ == '__main__':
    cli()



from .._group import project_group
from ...task.resource_maker import ResourceMaker
from ...task.cmd_task import CmdTask
from ...task_input.str_input import StrInput
from ...runner import runner
from ...helper.render_data.replacement_template import (
    Replacement
)

import os

# Common definitions

current_dir = os.path.dirname(__file__)

inputs = [
    StrInput(
        name='project-dir',
        prompt='Project directory',
        default='.'
    ),
    StrInput(
        name='project-name',
        prompt='Project name (can be empty)',
        default=''
    ),
]

replacements = Replacement().add_key_val(
    key='projectDir',
    value='input.project_dir'
).add_transformed_key_val(Replacement.ALL)(
    key='projectName',
    value=[
        'input.project_name',
        'os.path.basename(input.project_dir)'
    ]
).get()

# Task definitions

project_copy_resource = ResourceMaker(
    name='copy-project-resource',
    inputs=inputs,
    replacements=replacements,
    template_path=os.path.join(current_dir, 'project_template'),
    destination_path='{{input.project_dir}}',
    scaffold_locks=['{{input.project_dir}}/zrb_init.py']
)

project_create = CmdTask(
    name='create',
    group=project_group,
    upstreams=[project_copy_resource],
    inputs=[
        StrInput(name='project-dir', prompt='Project directory', default='.')
    ],
    cmd=[
        'set -e',
        'cd "{{input.project_dir}}"',
        'if [ ! -d .git ]',
        'then',
        '  echo Initialize project git repository',
        '  git init',
        'fi',
        'echo "Happy coding :)"',
    ]
)
runner.register(project_create)



from typing import Any
from ..task.task import Task
from ..task_input.str_input import StrInput
from ..runner import runner

# Common definitions


def _evaluate(*args: str, **kwargs: Any):
    expression: str = kwargs.get('expression', '')
    return eval(expression)


# Task definitions

evaluate = Task(
    name='eval',
    inputs=[
        StrInput(
            name='expression',
            shortcut='e',
            default='',
            description='Python expression',
        )
    ],
    runner=_evaluate,
    description='Evaluate Python expression',
    retry=0
)
runner.register(evaluate)



from zrb.builtin.project.project_create_task import project_create
import os
import pathlib
import shutil


def test_project_create():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    destination_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(destination_path):
        shutil.rmtree(destination_path)

    # first attempt should success
    first_attempt_loop = project_create.create_main_loop()
    result = first_attempt_loop(project_dir=destination_path)
    assert result

    # .git directory should exists
    assert os.path.isdir(os.path.join(destination_path, '.git'))
    # .gitignore file should exists
    assert os.path.isfile(os.path.join(destination_path, '.gitignore'))
    # zrb_init.py should exists
    assert os.path.isfile(os.path.join(destination_path, 'zrb_init.py'))

    # second attempt should success
    is_error = False
    try:
        second_attempt_loop = project_create.create_main_loop()
        result = second_attempt_loop(project_dir=destination_path)
    except Exception:
        is_error = True
    assert is_error



from ..dynamic_model import DynamicModel
from ..common import (
    coalesce, to_camel_case, to_pascal_case, to_kebab_case,
    to_snake_case, to_human_readable
)
import os
import datetime
import time

DEFAULT_RENDER_DATA = {
    'os': os,
    'datetime': datetime,
    'time': time,
    'zrb': DynamicModel(
        coalesce=coalesce,
        to_camel_case=to_camel_case,
        to_pascal_case=to_pascal_case,
        to_kebab_case=to_kebab_case,
        to_snake_case=to_snake_case,
        to_human_readable=to_human_readable
    )
}



from zrb.builtin.env_task import (
    env_show, env_get
)


def test_env_show():
    main_loop = env_show.create_main_loop()
    result = main_loop()
    assert result is None


def test_env_get():
    main_loop = env_get.create_main_loop()
    result = main_loop()
    assert len(result.keys()) > 0



from zrb.builtin.md5_task import md5_hash, md5_sum
import pathlib
import os


def test_md5_sum():
    dir_path = pathlib.Path(__file__).parent.absolute()
    main_loop = md5_sum.create_main_loop()
    result = main_loop(file=os.path.join(dir_path, 'md5_test_sum.txt'))
    assert result == '5f935c38df842258336f683502ac153d'


def test_md5_hash():
    main_loop = md5_hash.create_main_loop()
    result = main_loop(text='Philosopher Stone')
    assert result == '5f935c38df842258336f683502ac153d'



import libcst as cst


def add_import_module(code: str, module_path: str) -> str:
    module = cst.parse_module(code)

    # Create the new import statement
    new_import = cst.Import(
        names=[
            cst.ImportAlias(
                name=cst.parse_expression(module_path),
                asname=None
            )
        ],
    )

    last_import_index = None
    for i, node in enumerate(module.body):
        if isinstance(
            node, cst.SimpleStatementLine
        ) and isinstance(node.body[0], cst.Import):
            last_import_index = i

    # If there are no import statements
    # # add the new import at the beginning of the module
    if last_import_index is None:
        module.body.insert(0, cst.EmptyLine())
        module.body.insert(0, cst.EmptyLine())
        module.body.insert(0, new_import)
    else:
        module.body.insert(last_import_index + 1, cst.EmptyLine())
        module.body.insert(last_import_index + 1, new_import)

    # Generate the modified code
    modified_code = module.code
    return modified_code



from zrb.builtin.principle_task import (
    principle_show_dry, principle_show_kiss, principle_show_pancasila,
    principle_show_solid, principle_show_yagni
)


def test_principle_show_dry():
    main_loop = principle_show_dry.create_main_loop()
    result = main_loop()
    assert result is None


def test_principle_kiss_show():
    main_loop = principle_show_kiss.create_main_loop()
    result = main_loop()
    assert result is None


def test_principle_pancasila_show():
    main_loop = principle_show_pancasila.create_main_loop()
    result = main_loop()
    assert result is None


def test_principle_solid_show():
    main_loop = principle_show_solid.create_main_loop()
    result = main_loop()
    assert result is None


def test_principle_yagni_show():
    main_loop = principle_show_yagni.create_main_loop()
    result = main_loop()
    assert result is None



from typing import Any
from ._group import principle_show_group
from ..task.task import Task
from ..runner import runner

# Common definitions


def _principle_show_solid(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('S - Single Responsibility Principle')
    task.print_out('O - Open/Closed Principle')
    task.print_out('L - Liskov’s Substitution Principle')
    task.print_out('I - Interface Segregation Principle')
    task.print_out('D - Dependency Inversion Principle')


def _principle_show_yagni(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('Y - You')
    task.print_out('A - Aren\'t')
    task.print_out('G - Gonna')
    task.print_out('N - Need')
    task.print_out('I - It')


def _principle_show_dry(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('D - Don\'t')
    task.print_out('R - Repeat')
    task.print_out('Y - Yourself')


def _principle_show_kiss(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('K - Keep')
    task.print_out('I - It')
    task.print_out('S - Simple')
    task.print_out('S - Stupid')


def _principle_show_pancasila(*args: Any, **kwargs: Any):
    task: Task = kwargs['_task']
    task.print_out('1. Ketuhanan Yang Maha Esa')
    task.print_out('2. Kemanusiaan yang adil dan beradab')
    task.print_out('3. Persatuan Indonesia')
    task.print_out(' '.join([
        '4. Kerakyatan yang dipimpin oleh hikmat kebijaksanaan',
        'dalam permusyawaratan/perwakilan'
    ]))
    task.print_out('5. Keadilan sosial bagi seluruh rakyat Indonesia')


# Task definitions

principle_show_solid = Task(
    name='solid',
    group=principle_show_group,
    runner=_principle_show_solid,
    description='Show SOLID principle',
)
runner.register(principle_show_solid)

principle_show_yagni = Task(
    name='yagni',
    group=principle_show_group,
    runner=_principle_show_yagni,
    description='Show YAGNI principle',
)
runner.register(principle_show_yagni)

principle_show_dry = Task(
    name='dry',
    group=principle_show_group,
    runner=_principle_show_dry,
    description='Show dry principle',
)
runner.register(principle_show_dry)

principle_show_kiss = Task(
    name='kiss',
    group=principle_show_group,
    runner=_principle_show_kiss,
    description='Show kiss principle',
)
runner.register(principle_show_kiss)

principle_show_pancasila = Task(
    name='pancasila',
    group=principle_show_group,
    runner=_principle_show_pancasila,
    description='Show pancasila',
)
runner.register(principle_show_pancasila)



from ..runner import runner
from .loader.load_module import load_module
from ..config.config import init_scripts, should_load_builtin
from .log import logger
from .accessories.color import colored

import click
import os


def create_cli() -> click.Group:
    cli = click.Group(name='zrb', help='Your faithful companion.')

    if 'PYTHONUNBUFFERED' not in os.environ:
        os.environ['PYTHONUNBUFFERED'] = '1'

    # load from ZRB_INIT_SCRIPTS environment
    for init_script in init_scripts:
        load_module(script_path=init_script)

    # Load default tasks
    if should_load_builtin:
        logger.info(colored('Loading builtins', attrs=['dark']))
        from .. import builtin
        assert builtin

    # Load zrb_init.py
    project_dir = os.getcwd()
    load_zrb_init(project_dir)

    # Serve all tasks registered to runner
    cli = runner.serve(cli)
    return cli


def load_zrb_init(project_dir: str):
    project_script = os.path.join(project_dir, 'zrb_init.py')
    if os.path.isfile(project_script):
        load_module(script_path=project_script)
        os.environ['ZRB_PROJECT_DIR'] = project_dir
        logger.info(colored(f'ZRB_PROJECT_DIR={project_dir}', attrs=['dark']))
        add_project_dir_to_python_path(project_dir)
        python_path = os.getenv('PYTHONPATH')
        logger.info(colored(f'PYTHONPATH={python_path}', attrs=['dark']))
        return
    # zrb_init.py not found, look for it on the parent directory
    parent_project_dir = os.path.dirname(project_dir)
    if parent_project_dir == project_dir:
        # giving up, already on the root level directory
        logger.debug('Cannot find zrb_init.py')
        return
    # do the same thing again
    load_zrb_init(parent_project_dir)


def add_project_dir_to_python_path(project_dir: str):
    current_python_path = os.getenv('PYTHONPATH')
    if current_python_path is None or current_python_path == '':
        os.environ['PYTHONPATH'] = project_dir
        return
    os.environ['PYTHONPATH'] = ':'.join([
        current_python_path, project_dir
    ])



from typing import Any
from ._group import base64_group
from ..task.task import Task
from ..task_input.str_input import StrInput
from ..runner import runner

import base64

# Common definitions

inputs = [
    StrInput(
        name='text',
        shortcut='t',
        description='Text',
        default=''
    )
]


def _base64_encode(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    encoded_text = base64.b64encode(text.encode())
    return encoded_text.decode()


def _base64_decode(*args: str, **kwargs: Any):
    text: str = kwargs.get('text', '')
    encoded_text = base64.b64decode(text.encode())
    return encoded_text.decode()


# Task definitions

base64_encode = Task(
    name='encode',
    group=base64_group,
    inputs=inputs,
    runner=_base64_encode,
    description='Encode base64 task',
    retry=0
)
runner.register(base64_encode)

base64_decode = Task(
    name='decode',
    group=base64_group,
    inputs=inputs,
    runner=_base64_decode,
    description='Decode base64 task',
    retry=0
)
runner.register(base64_decode)



from ._group import ubuntu_group, ubuntu_install_group
from ..task.cmd_task import CmdTask
from ..runner import runner


ubuntu_update = CmdTask(
    name='ubuntu',
    group=ubuntu_group,
    description='Update ubuntu',
    cmd=[
        'sudo apt-get update',
        'sudo apt-get upgrade -y',
    ],
    checking_interval=3
)
runner.register(ubuntu_update)

ubuntu_install_toys = CmdTask(
    name='toys',
    group=ubuntu_install_group,
    description='Install ubuntu toy packages',
    cmd=[
        'sudo apt-get install -y lolcat cowsay figlet neofetch',
    ],
    upstreams=[ubuntu_update],
    checking_interval=3
)
runner.register(ubuntu_install_toys)

ubuntu_install_packages = CmdTask(
    name='packages',
    group=ubuntu_install_group,
    description='Install ubuntu packages',
    cmd=[
        'sudo apt-get install -y \\',
        'build-essential python3-distutils libssl-dev zlib1g-dev \\'
        'libbz2-dev libreadline-dev libsqlite3-dev libpq-dev python3-dev \\',
        'llvm libncurses5-dev libncursesw5-dev xz-utils tk-dev libffi-dev \\',
        'liblzma-dev python3-openssl bison libblas-dev liblapack-dev \\',
        'gfortran rustc fd-find ripgrep wget curl git ncat cmake make tree \\',
        'tmux zsh neovim xdotool xsel'
    ],
    upstreams=[ubuntu_update],
    checking_interval=3
)
runner.register(ubuntu_install_packages)



from typing import Any, Callable
from .base_task import BaseTask
from typeguard import typechecked


@typechecked
class Task(BaseTask):
    '''
    Common Task.
    Exactly the same as BaseTask
    '''

    def runner(self, runner: Callable[..., Any]):
        if self._runner is not None:
            raise Exception('Cannot set multiple runners for a single task.')
        self._runner = runner



from zrb.builtin.base64_task import base64_encode, base64_decode


def test_base64_encode():
    main_loop = base64_encode.create_main_loop()
    result = main_loop(text='Philosopher Stone')
    assert result == 'UGhpbG9zb3BoZXIgU3RvbmU='


def test_base64_decode():
    main_loop = base64_decode.create_main_loop()
    result = main_loop(text='UGhpbG9zb3BoZXIgU3RvbmU=')
    assert result == 'Philosopher Stone'



from ..task_group.group import Group

project_group = Group(
    name='project', description='Project related commands'
)
task_group = Group(
    name='task', description='Task related commands'
)
env_group = Group(
    name='env', description='Env related commands'
)
ubuntu_group = Group(
    name='ubuntu', description='Ubuntu related commands'
)
ubuntu_install_group = Group(
    name='install', description='Install things on ubuntu', parent=ubuntu_group
)
principle_group = Group(
    name='principle', description='Principle related commands'
)
principle_show_group = Group(
    name='show', description='Showing principles', parent=principle_group
)
base64_group = Group(
    name='base64', description='Base64 related commands'
)
md5_group = Group(
    name='md5', description='MD5 related commands'
)




from typing import Any


class DynamicModel:
    def __init__(self, **kwargs: Any):
        self.__dict__.update(kwargs)

    def __repr__(self) -> str:
        return str(self.__dict__)

    def __getattribute__(self, name):
        '''
        Return None if attribute not found
        '''
        try:
            return super().__getattribute__(name)
        except AttributeError:
            return None



from typing import Any, Callable, List, Mapping, Optional
from typeguard import typechecked
from .base_task import BaseTask
from ..task_env.env import Env
from ..task_group.group import Group
from ..task_input.base_input import BaseInput
from ..helper.file.copy_tree import copy_tree

import os


@typechecked
class ResourceMaker(BaseTask):

    def __init__(
        self,
        name: str,
        template_path: str,
        destination_path: str,
        replacements: Mapping[str, str] = {},
        excludes: List[str] = [],
        group: Optional[Group] = None,
        inputs: List[BaseInput] = [],
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = '',
        upstreams: List[BaseTask] = [],
        scaffold_locks: List[str] = []
    ):
        BaseTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            checkers=[],
            checking_interval=0.1,
            retry=0,
            retry_interval=0
        )
        self.template_path = template_path
        self.destination_path = destination_path
        self.excludes = excludes
        self.replacements = replacements
        self.scaffold_locks = scaffold_locks

    def create_main_loop(
        self, env_prefix: str = '', raise_error: bool = True
    ) -> Callable[..., bool]:
        return super().create_main_loop(env_prefix, raise_error)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        # render parameters
        template_path = self.render_str(self.template_path)
        destination_path = self.render_str(self.destination_path)
        self.log_debug(f'Render excludes: {self.excludes}')
        excludes = [
            self.render_str(exclude)
            for exclude in self.excludes
        ]
        self.log_debug(f'Rendered excludes: {excludes}')
        self.log_debug(f'Render replacements: {self.replacements}')
        replacements: Mapping[str, str] = {
            old: self.render_str(new)
            for old, new in self.replacements.items()
        }
        self.log_debug(f'Rendered replacements: {replacements}')
        # check scaffold locks
        for scaffold_lock in self.scaffold_locks:
            self.log_debug(f'Render scaaffold lock: {scaffold_lock}')
            rendered_scaffold_lock = self.render_str(scaffold_lock)
            self.log_debug(f'Rendered scaffold lock: {rendered_scaffold_lock}')
            if not os.path.exists(rendered_scaffold_lock):
                continue
            raise Exception(' '.join([
                'Operation cancelled since resource already exists:',
                f'{rendered_scaffold_lock},',
            ]))
        self.print_out('    \n'.join([
            f'Create resource: {destination_path}',
            f'Template: {template_path}',
            f'Replacements: {replacements}',
            f'Excludes: {excludes}',
        ]))
        copy_tree(
            src=template_path,
            dst=destination_path,
            replacements=replacements,
            excludes=excludes
        )
        return True



from zrb.builtin.project.project_create_task import project_create
from zrb.builtin.task.task_create_task import task_create
import os
import pathlib
import shutil


def test_task_create():
    # prepare path
    dir_path = pathlib.Path(__file__).parent.absolute()
    project_path = os.path.join(dir_path, 'app')
    # remove destination if exists
    if os.path.exists(project_path):
        shutil.rmtree(project_path)

    create_project = project_create.create_main_loop()
    create_project(project_dir=project_path)

    destination_path = os.path.join(project_path, 'transmutate')

    # first attempt should success
    first_attempt_loop = task_create.create_main_loop()
    result = first_attempt_loop(task_dir=destination_path)
    assert result

    # transmutation directory should exists
    assert os.path.isdir(os.path.join(destination_path))
    # transmutation.py file should exists
    assert os.path.isfile(
        os.path.join(destination_path, 'transmutate.py')
    )

    # second attempt should success
    is_error = False
    try:
        second_attempt_loop = task_create.create_main_loop()
        result = second_attempt_loop(task_dir=destination_path)
    except Exception:
        is_error = True
    assert is_error



from typing import Any, List, Mapping, Optional, Union
from typeguard import typechecked
from ..helper.accessories.color import (
    get_random_color, is_valid_color, colored
)
from ..helper.accessories.icon import get_random_icon
from ..helper.keyval.get_object_from_keyval import get_object_from_keyval
from ..helper.string.conversion import (
    to_cmd_name, to_variable_name
)
from ..helper.render_data.default_render_data import DEFAULT_RENDER_DATA
from ..helper.log import logger
from ..task_env.env import Env
from ..task_group.group import Group

import datetime
import os
import sys
import time
import jinja2


MAX_NAME_LENGTH = 20
MULTILINE_INDENT = ' ' * 8


class AnyExtensionFileSystemLoader(jinja2.FileSystemLoader):
    def get_source(self, environment, template):
        for search_dir in self.searchpath:
            file_path = os.path.join(search_dir, template)
            if os.path.exists(file_path):
                with open(file_path, 'r') as file:
                    contents = file.read()
                    return contents, file_path, lambda: False
        raise jinja2.TemplateNotFound(template)


@typechecked
class TimeTracker():

    def __init__(self):
        self._start_time: float = 0
        self._end_time: float = 0

    def reset_timer(self):
        self._start_time = 0
        self._end_time = 0

    def start_timer(self):
        self._start_time = time.time()

    def end_timer(self):
        self._end_time = time.time()

    def get_elapsed_time(self) -> float:
        return self._end_time - self._start_time


@typechecked
class AttemptTracker():

    def __init__(self, retry: int = 2):
        self.retry = retry
        self._attempt: int = 1

    def get_max_attempt(self) -> int:
        return self.retry + 1

    def get_attempt(self) -> int:
        return self._attempt

    def increase_attempt(self):
        self._attempt += 1

    def reset_attempt(self):
        self._attempt = 0

    def should_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt <= max_attempt

    def is_last_attempt(self) -> bool:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        return attempt >= max_attempt


@typechecked
class FinishTracker():

    def __init__(self):
        self._is_done: bool = False

    def mark_as_done(self):
        self._is_done = True

    def mark_as_undone(self):
        self._is_done = False

    def is_done(self) -> bool:
        return self._is_done


@typechecked
class PidModel():

    def __init__(self):
        self.zrb_task_pid: int = os.getpid()

    def set_task_pid(self, pid: int):
        self.zrb_task_pid = pid

    def get_task_pid(self) -> int:
        return self.zrb_task_pid


@typechecked
class TaskDataModel():

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
    ):
        self.name = name
        self.group = group
        self.envs = envs
        for key in os.environ:
            self.envs.append(Env(name=key, os_name=key))
        self.icon = icon
        self.color = color
        self._input_map: Mapping[str, Any] = {}
        self._env_map: Mapping[str, str] = {}
        self._is_keyval_set = False  # Flag

    def get_icon(self) -> str:
        if self.icon is None or self.icon == '':
            self.icon = get_random_icon()
        return self.icon

    def get_color(self) -> str:
        if self.color is None or not is_valid_color(self.color):
            self.color = get_random_color()
        return self.color

    def get_cmd_name(self) -> str:
        return to_cmd_name(self.name)

    def log_debug(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.debug(colored_message)

    def log_warn(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.warning(colored_message)

    def log_info(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', attrs=['dark']
        )
        logger.info(colored_message)

    def log_error(self, message: Any):
        prefix = self._get_log_prefix()
        colored_message = colored(
            f'{prefix} • {message}', color='red', attrs=['bold']
        )
        logger.error(colored_message, exc_info=True)

    def print_out(self, msg: Any):
        prefix = self._get_colored_print_prefix()
        print(f'🤖 ➜  {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def print_err(self, msg: Any):
        prefix = self._get_colored_print_prefix()
        print(f'🤖 ⚠  {prefix} • {msg}'.rstrip(), file=sys.stderr)

    def colored(self, text: str) -> str:
        return colored(text, color=self.get_color())

    def play_bell(self):
        print('\a')

    def _get_colored_print_prefix(self) -> str:
        return self.colored(self._get_print_prefix())

    def _get_print_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=True)
        icon = self.get_icon()
        truncated_name = self._get_truncated_complete_name()
        return f'{common_prefix} • {icon} {truncated_name}'

    def _get_log_prefix(self) -> str:
        common_prefix = self._get_common_prefix(show_time=False)
        icon = self.get_icon()
        filled_name = self._get_filled_complete_name()
        return f'{common_prefix} • {icon} {filled_name}'

    def _get_common_prefix(self, show_time: bool) -> str:
        attempt = self.get_attempt()
        max_attempt = self.get_max_attempt()
        pid = self.get_task_pid()
        if show_time:
            now = datetime.datetime.now().isoformat()
            return f'{now} ⚙ {pid} ➤ {attempt} of {max_attempt}'
        return f'⚙ {pid} ➤ {attempt} of {max_attempt}'

    def _get_filled_complete_name(self) -> str:
        return self._get_complete_name().rjust(MAX_NAME_LENGTH, ' ')

    def _get_truncated_complete_name(self):
        filled_name = self._get_filled_complete_name()
        if len(filled_name) > MAX_NAME_LENGTH:
            return '..' + filled_name[-(MAX_NAME_LENGTH-2):]
        return filled_name

    def get_input_map(self) -> Mapping[str, Any]:
        return self._input_map

    def get_env_map(self) -> Mapping[str, str]:
        return self._env_map

    def inject_env_map(
        self, env_map: Mapping[str, str], override: bool = False
    ):
        for key, val in env_map.items():
            if override or key not in self._env_map:
                self._env_map[key] = val

    def get_float(self, val: Union[str, float]) -> float:
        if isinstance(val, str):
            return float(self.render_str(val))
        return val

    def get_int(self, val: Union[str, int]) -> int:
        if isinstance(val, str):
            return int(self.render_str(val))
        return val

    def get_bool(self, val: Union[str, bool]) -> bool:
        if isinstance(val, str):
            str_val = self.render_str(val)
            if str_val.lower() in ['true', '1', 'yes', 'y', 'active']:
                return True
            if str_val.lower() in ['false', '0', 'no', 'n', 'inactive']:
                return False
            raise Exception(f'Cannot infer boolean value from {str_val}')
        return val

    def render_str(self, val: str) -> str:
        template = jinja2.Template(val)
        data = self._get_render_data()
        self.log_debug('\n'.join([
            'Render string template:',
            self._get_multiline_repr(val),
            'With data:',
            self._get_map_repr(data)
        ]))
        rendered_text = template.render(data)
        self.log_debug('\n'.join([
            'Get rendered result:',
            self._get_multiline_repr(rendered_text),
        ]))
        return rendered_text

    def render_file(self, location: str) -> str:
        location_dir = os.path.dirname(location)
        env = jinja2.Environment(
            loader=AnyExtensionFileSystemLoader([location_dir])
        )
        template = env.get_template(location)
        data = self._get_render_data()
        data['TEMPLATE_DIR'] = location_dir
        self.log_debug('\n'.join([
            f'Render template file: {template}',
            'With data:',
            self._get_map_repr(data)
        ]))
        rendered_text = template.render(data)
        self.log_debug('\n'.join([
            'Get rendered result:',
            self._get_multiline_repr(rendered_text),
        ]))
        return rendered_text

    def _get_render_data(self) -> Mapping[str, Any]:
        render_data = dict(DEFAULT_RENDER_DATA)
        render_data.update({
            'env': get_object_from_keyval(self._env_map),
            'input': get_object_from_keyval(self._input_map),
        })
        return render_data

    def _get_multiline_repr(self, text: str) -> str:
        lines_repr: List[str] = []
        for index, line in enumerate(text.split('\n')):
            line_number_repr = str(index + 1).rjust(4, '0')
            lines_repr.append(f'{MULTILINE_INDENT}{line_number_repr} | {line}')
        return '\n'.join(lines_repr)

    def _get_map_repr(self, data: Mapping[str, Any]) -> str:
        lines_repr: List[str] = []
        for key, val in data.items():
            lines_repr.append(f'{MULTILINE_INDENT}{key}: {val}')
        return '\n'.join(lines_repr)

    def set_local_keyval(
        self,
        input_map: Mapping[str, Any],
        env_prefix: str = ''
    ):
        if self._is_keyval_set:
            return True
        self._is_keyval_set = True
        self._input_map: Mapping[str, Any]
        for input_name, val in input_map.items():
            self._input_map[to_variable_name(input_name)] = val
        self.log_debug('\n'.join([
            'Set input map:',
            self._get_map_repr(self._input_map)
        ]))
        self._env_map = os.environ
        for task_env in self.envs:
            env_name = task_env.name
            env_value = task_env.get(env_prefix)
            self._env_map[env_name] = env_value
        self.log_debug('\n'.join([
            'Set env map:',
            self._get_map_repr(self._env_map)
        ]))

    def _get_complete_name(self) -> str:
        cmd_name = self.get_cmd_name()
        if self.group is None:
            return cmd_name
        group_cmd_name = self.group.get_complete_name()
        return f'{group_cmd_name} {cmd_name}'

    def show_celebration(self):
        complete_name = self._get_complete_name()
        elapsed_time = self.get_elapsed_time()
        icon = self.get_icon()
        message = '\n'.join([
            '🤖 🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉',
            f'🤖 {icon} {complete_name} completed in',
            f'🤖 {icon} {elapsed_time} seconds',
            '🤖 🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉🎉'
        ])
        print(message, file=sys.stderr)
        self.play_bell()


@typechecked
class TaskModel(
    TaskDataModel, PidModel, FinishTracker, AttemptTracker, TimeTracker
):

    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        envs: List[Env] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        retry: int = 2,
    ):
        TaskDataModel.__init__(
            self,
            name=name,
            group=group,
            envs=envs,
            icon=icon,
            color=color
        )
        retry = self.ensure_non_negative(retry, 'Find negative retry')
        PidModel.__init__(self)
        FinishTracker.__init__(self)
        AttemptTracker.__init__(self, retry=retry)
        TimeTracker.__init__(self)

    def ensure_non_negative(self, value: float, error_label: str) -> float:
        if value < 0:
            self.log_warn(
                f'{error_label}: {value}'
            )
            return 0
        return value



from typing import List, Mapping, Optional, TypeVar, Union
from ..common import to_pascal_case


def tpl(val: str) -> str:
    return '{{' + val + '}}'


def coalesce(val: str, *alternatives: str) -> str:
    values = ', '.join([val] + list(alternatives))
    return f'zrb.coalesce({values})'


def camel(val: str) -> str:
    return f'zrb.to_camel_case({val})'


def kebab(val: str) -> str:
    return f'zrb.to_kebab_case({val})'


def snake(val: str) -> str:
    return f'zrb.to_snake_case({val})'


def pascal(val: str) -> str:
    return f'zrb.to_pascal_case({val})'


def human(val: str) -> str:
    return f'zrb.to_human_readable({val})'


TReplacement = TypeVar('TReplacement', bound='Replacement')


class Replacement():
    CAMEL = 'camel'
    PASCAL = 'pascal'
    KEBAB = 'kebab'
    SNAKE = 'snake'
    HUMAN = 'human'
    ALL = ['camel', 'pascal', 'kebab', 'snake', 'human']

    def __init__(self, replacements: Optional[Mapping[str, str]] = None):
        self.replacements = replacements if replacements is not None else {}
        self.tpl_map = {
            Replacement.CAMEL: camel,
            Replacement.PASCAL: pascal,
            Replacement.KEBAB: kebab,
            Replacement.SNAKE: snake,
            Replacement.HUMAN: human,
        }

    def get(self) -> Mapping[str, str]:
        return self.replacements

    def add_key_val(
        self, key: str, value: Union[str, List[str]]
    ) -> TReplacement:
        val = self._get_val(value)
        self.replacements[key] = tpl(val)
        return self

    def _get_val(self, value: Union[str, List[str]]) -> str:
        if isinstance(value, str):
            return value
        return coalesce(*value)

    def add_transformed_key_val(self, transformations: List[str]):
        def add_key_val(
            key: str, value: Union[str, List[str]]
        ) -> TReplacement:
            return self._add_transformed_key_val(
                transformations=transformations,
                key=key,
                value=value
            )
        return add_key_val

    def _add_transformed_key_val(
        self,
        transformations: List[str],
        key: str,
        value: Union[str, List[str]]
    ) -> TReplacement:
        for trans_name in transformations:
            if trans_name not in self.tpl_map:
                raise Exception(f'Invalid transformation {trans_name}')
            tpl_transformer = self.tpl_map[trans_name]
            new_key = to_pascal_case(f'{trans_name} {key}')
            new_val = tpl_transformer(self._get_val(value))
            self.add_key_val(new_key, new_val)
        self.add_key_val(key, value)
        return self



from . import base64_task
from . import env_task
from . import eval_task
from . import md5_task
from . import principle_task
from . import ubuntu_task
from .project import project_create_task
from .task import task_create_task

assert base64_task
assert env_task
assert eval_task
assert md5_task
assert principle_task
assert ubuntu_task
assert project_create_task
assert task_create_task



from zrb import (
    CmdTask, runner
)


SnakeTaskName = CmdTask(
    name='KebabTaskName',
    description='HumanTaskName',
    cmd=[
        'echo HumanTaskName'
    ]
)
runner.register(SnakeTaskName)



from .runner import runner
from .helper.loader.load_module import load_module
from .task.task import Task
from .task.cmd_task import CmdTask
from .task.http_checker import HTTPChecker
from .task.port_checker import PortChecker
from .task.resource_maker import ResourceMaker
from .task_input.str_input import StrInput
from .task_env.env import Env
from .task_group.group import Group
from .builtin import _group as builtin_group


assert runner
assert load_module
assert Task
assert CmdTask
assert HTTPChecker
assert PortChecker
assert ResourceMaker
assert StrInput
assert Env
assert Group
assert builtin_group
